@@ -1,7 +1,15 @@
+# -*- coding: utf-8 -*-
 # Copyright (c) 2009-2011, 2013-2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
-# Copyright (c) 2013-2016 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2010 Daniel Harding <dharding@gmail.com>
+# Copyright (c) 2013-2016, 2018 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2013-2014 Google, Inc.
-# Copyright (c) 2015-2016 Cara Vinson <ceridwenv@gmail.com>
+# Copyright (c) 2015-2016 Ceridwen <ceridwenv@gmail.com>
+# Copyright (c) 2016 Jared Garst <jgarst@users.noreply.github.com>
+# Copyright (c) 2016 Jakub Wilk <jwilk@jwilk.net>
+# Copyright (c) 2017 Łukasz Rogalski <rogalski.91@gmail.com>
+# Copyright (c) 2017 rr- <rr-@sakuya.pl>
+# Copyright (c) 2018 brendanator <brendan.maginnis@gmail.com>
+# Copyright (c) 2018 Nick Drozd <nicholasdrozd@gmail.com>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -15,13 +23,13 @@
 """
 import sys
 
-import six
-
 
 # pylint: disable=unused-argument
 
+DOC_NEWLINE = '\0'
 
-class AsStringVisitor(object):
+
+class AsStringVisitor:
     """Visitor to render an Astroid node as a valid python code string"""
 
     def __init__(self, indent):
@@ -29,12 +37,47 @@ class AsStringVisitor(object):
 
     def __call__(self, node):
         """Makes this visitor behave as a simple function"""
-        return node.accept(self)
+        return node.accept(self).replace(DOC_NEWLINE, '\n')
 
-    def _stmt_list(self, stmts):
+    def _docs_dedent(self, doc):
+        """Stop newlines in docs being indented by self._stmt_list"""
+        return '\n%s"""%s"""' % (self.indent, doc.replace('\n', DOC_NEWLINE))
+
+    def _stmt_list(self, stmts, indent=True):
         """return a list of nodes to string"""
-        stmts = '\n'.join([nstr for nstr in [n.accept(self) for n in stmts] if nstr])
-        return self.indent + stmts.replace('\n', '\n'+self.indent)
+        stmts = '\n'.join(
+            [nstr for nstr in [n.accept(self) for n in stmts] if nstr])
+        if indent:
+            return self.indent + stmts.replace('\n', '\n' + self.indent)
+
+        return stmts
+
+    def _precedence_parens(self, node, child, is_left=True):
+        """Wrap child in parens only if required to keep same semantics"""
+        if self._should_wrap(node, child, is_left):
+            return '(%s)' % child.accept(self)
+
+        return child.accept(self)
+
+    def _should_wrap(self, node, child, is_left):
+        """Wrap child if:
+            - it has lower precedence
+            - same precedence with position opposite to associativity direction
+        """
+        node_precedence = node.op_precedence()
+        child_precedence = child.op_precedence()
+
+        if node_precedence > child_precedence:
+            # 3 * (4 + 5)
+            return True
+
+        if (node_precedence == child_precedence and
+                is_left != node.op_left_associative()):
+            # 3 - (4 - 5)
+            # (2**3)**4
+            return True
+
+        return False
 
 
     ## visit_<node> methods ###########################################
@@ -67,18 +110,32 @@ class AsStringVisitor(object):
         """return an astroid.AugAssign node as string"""
         return '%s %s %s' % (node.target.accept(self), node.op, node.value.accept(self))
 
+    def visit_annassign(self, node):
+        """Return an astroid.AugAssign node as string"""
+
+        target = node.target.accept(self)
+        annotation = node.annotation.accept(self)
+        if node.value is None:
+            return '%s: %s' % (target, annotation)
+        return '%s: %s = %s' % (target, annotation, node.value.accept(self))
+
     def visit_repr(self, node):
         """return an astroid.Repr node as string"""
         return '`%s`' % node.value.accept(self)
 
     def visit_binop(self, node):
         """return an astroid.BinOp node as string"""
-        return '(%s) %s (%s)' % (node.left.accept(self), node.op, node.right.accept(self))
+        left = self._precedence_parens(node, node.left)
+        right = self._precedence_parens(node, node.right, is_left=False)
+        if node.op == '**':
+            return '%s%s%s' % (left, node.op, right)
+
+        return '%s %s %s' % (left, node.op, right)
 
     def visit_boolop(self, node):
         """return an astroid.BoolOp node as string"""
-        return (' %s ' % node.op).join(['(%s)' % n.accept(self)
-                                        for n in node.values])
+        values = ['%s' % self._precedence_parens(node, n) for n in node.values]
+        return (' %s ' % node.op).join(values)
 
     def visit_break(self, node):
         """return an astroid.Break node as string"""
@@ -86,7 +143,7 @@ class AsStringVisitor(object):
 
     def visit_call(self, node):
         """return an astroid.Call node as string"""
-        expr_str = node.func.accept(self)
+        expr_str = self._precedence_parens(node, node.func)
         args = [arg.accept(self) for arg in node.args]
         if node.keywords:
             keywords = [kwarg.accept(self) for kwarg in node.keywords]
@@ -111,15 +168,17 @@ class AsStringVisitor(object):
                     bases = '(metaclass=%s)' % metaclass.name
             else:
                 bases = '(%s)' % bases if bases else ''
-        docs = '\n%s"""%s"""' % (self.indent, node.doc) if node.doc else ''
+        docs = self._docs_dedent(node.doc) if node.doc else ''
         return '\n\n%sclass %s%s:%s\n%s\n' % (decorate, node.name, bases, docs,
                                               self._stmt_list(node.body))
 
     def visit_compare(self, node):
         """return an astroid.Compare node as string"""
-        rhs_str = ' '.join(['%s %s' % (op, expr.accept(self))
-                            for op, expr in node.ops])
-        return '%s %s' % (node.left.accept(self), rhs_str)
+        rhs_str = ' '.join([
+            '%s %s' % (op, self._precedence_parens(node, expr, is_left=False))
+            for op, expr in node.ops
+        ])
+        return '%s %s' % (self._precedence_parens(node, node.left), rhs_str)
 
     def visit_comprehension(self, node):
         """return an astroid.Comprehension node as string"""
@@ -233,13 +292,11 @@ class AsStringVisitor(object):
     def visit_functiondef(self, node):
         """return an astroid.Function node as string"""
         decorate = node.decorators.accept(self) if node.decorators else ''
-        docs = '\n%s"""%s"""' % (self.indent, node.doc) if node.doc else ''
-        return_annotation = ''
-        if six.PY3 and node.returns:
+        docs = self._docs_dedent(node.doc) if node.doc else ''
+        trailer = ':'
+        if node.returns:
             return_annotation = '->' + node.returns.as_string()
             trailer = return_annotation + ":"
-        else:
-            trailer = ":"
         def_format = "\n%sdef %s(%s)%s%s\n%s"
         return def_format % (decorate, node.name,
                              node.args.accept(self),
@@ -253,7 +310,7 @@ class AsStringVisitor(object):
 
     def visit_attribute(self, node):
         """return an astroid.Getattr node as string"""
-        return '%s.%s' % (node.expr.accept(self), node.attrname)
+        return '%s.%s' % (self._precedence_parens(node, node.expr), node.attrname)
 
     def visit_global(self, node):
         """return an astroid.Global node as string"""
@@ -262,15 +319,18 @@ class AsStringVisitor(object):
     def visit_if(self, node):
         """return an astroid.If node as string"""
         ifs = ['if %s:\n%s' % (node.test.accept(self), self._stmt_list(node.body))]
-        if node.orelse:# XXX use elif ???
+        if node.has_elif_block():
+            ifs.append('el%s' % self._stmt_list(node.orelse, indent=False))
+        elif node.orelse:
             ifs.append('else:\n%s' % self._stmt_list(node.orelse))
         return '\n'.join(ifs)
 
     def visit_ifexp(self, node):
         """return an astroid.IfExp node as string"""
-        return '%s if %s else %s' % (node.body.accept(self),
-                                     node.test.accept(self),
-                                     node.orelse.accept(self))
+        return '%s if %s else %s' % (
+            self._precedence_parens(node, node.body, is_left=True),
+            self._precedence_parens(node, node.test, is_left=True),
+            self._precedence_parens(node, node.orelse, is_left=False))
 
     def visit_import(self, node):
         """return an astroid.Import node as string"""
@@ -284,8 +344,12 @@ class AsStringVisitor(object):
 
     def visit_lambda(self, node):
         """return an astroid.Lambda node as string"""
-        return 'lambda %s: %s' % (node.args.accept(self),
-                                  node.body.accept(self))
+        args = node.args.accept(self)
+        body = node.body.accept(self)
+        if args:
+            return 'lambda %s: %s' % (args, body)
+
+        return 'lambda: %s' % body
 
     def visit_list(self, node):
         """return an astroid.List node as string"""
@@ -333,13 +397,17 @@ class AsStringVisitor(object):
 
     def visit_return(self, node):
         """return an astroid.Return node as string"""
+        if node.is_tuple_return() and len(node.value.elts) > 1:
+            elts = [child.accept(self) for child in node.value.elts]
+            return 'return %s' % ', '.join(elts)
+
         if node.value:
             return 'return %s' % node.value.accept(self)
 
         return 'return'
 
     def visit_index(self, node):
-        """return a astroid.Index node as string"""
+        """return an astroid.Index node as string"""
         return node.value.accept(self)
 
     def visit_set(self, node):
@@ -352,17 +420,18 @@ class AsStringVisitor(object):
                             ' '.join([n.accept(self) for n in node.generators]))
 
     def visit_slice(self, node):
-        """return a astroid.Slice node as string"""
+        """return an astroid.Slice node as string"""
         lower = node.lower.accept(self) if node.lower else ''
-        upper = node.upper.accept(self) if node.upper else''
+        upper = node.upper.accept(self) if node.upper else ''
         step = node.step.accept(self) if node.step else ''
         if step:
             return '%s:%s:%s' % (lower, upper, step)
-        return  '%s:%s' % (lower, upper)
+        return '%s:%s' % (lower, upper)
 
     def visit_subscript(self, node):
         """return an astroid.Subscript node as string"""
-        return '%s[%s]' % (node.value.accept(self), node.slice.accept(self))
+        return '%s[%s]' % (self._precedence_parens(node, node.value),
+                           node.slice.accept(self))
 
     def visit_tryexcept(self, node):
         """return an astroid.TryExcept node as string"""
@@ -390,7 +459,7 @@ class AsStringVisitor(object):
             operator = 'not '
         else:
             operator = node.op
-        return '%s%s' % (operator, node.operand.accept(self))
+        return '%s%s' % (operator, self._precedence_parens(node, node.operand))
 
     def visit_while(self, node):
         """return an astroid.While node as string"""
@@ -402,8 +471,8 @@ class AsStringVisitor(object):
 
     def visit_with(self, node): # 'with' without 'as' is possible
         """return an astroid.With node as string"""
-        items = ', '.join(('(%s)' % expr.accept(self)) +
-                          (vars and ' as (%s)' % (vars.accept(self)) or '')
+        items = ', '.join(('%s' % expr.accept(self)) +
+                          (vars and ' as %s' % (vars.accept(self)) or '')
                           for expr, vars in node.items)
         return 'with %s:\n%s' % (items, self._stmt_list(node.body))
 
@@ -494,6 +563,11 @@ class AsStringVisitor3(AsStringVisitor):
     def visit_formattedvalue(self, node):
         return '{%s}' % node.value.accept(self)
 
+    def visit_comprehension(self, node):
+        """return an astroid.Comprehension node as string"""
+        return '%s%s' % ('async ' if node.is_async else '',
+                         super(AsStringVisitor3, self).visit_comprehension(node))
+
 
 def _import_string(names):
     """return a list of (name, asname) formatted as a string"""
