@@ -1,7 +1,15 @@
 # Copyright (c) 2009-2011, 2013-2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
-# Copyright (c) 2014-2016 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2012 FELD Boris <lothiraldan@gmail.com>
+# Copyright (c) 2014-2018 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2014 Google, Inc.
-# Copyright (c) 2015-2016 Cara Vinson <ceridwenv@gmail.com>
+# Copyright (c) 2014 Eevee (Alex Munroe) <amunroe@yelp.com>
+# Copyright (c) 2015-2016 Ceridwen <ceridwenv@gmail.com>
+# Copyright (c) 2015 Florian Bruhin <me@the-compiler.org>
+# Copyright (c) 2016-2017 Derek Gustafson <degustaf@gmail.com>
+# Copyright (c) 2017 Calen Pennington <calen.pennington@gmail.com>
+# Copyright (c) 2018 Bryce Guinta <bryce.paul.guinta@gmail.com>
+# Copyright (c) 2018 Nick Drozd <nicholasdrozd@gmail.com>
+# Copyright (c) 2018 Daniel Colascione <dancol@dancol.org>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -10,21 +18,22 @@
 inference utils.
 """
 
+import builtins
 import collections
 import sys
 
-import six
-
 from astroid import context as contextmod
 from astroid import exceptions
 from astroid import util
 
 objectmodel = util.lazy_import('interpreter.objectmodel')
-BUILTINS = six.moves.builtins.__name__
+helpers = util.lazy_import('helpers')
+BUILTINS = builtins.__name__
 manager = util.lazy_import('manager')
 MANAGER = manager.AstroidManager()
 
 if sys.version_info >= (3, 0):
+    # TODO: check if needs special treatment
     BUILTINS = 'builtins'
     BOOL_SPECIAL_METHOD = '__bool__'
 else:
@@ -39,11 +48,10 @@ PROPERTIES = {BUILTINS + '.property', 'abc.abstractproperty'}
 # Also, these aren't qualified, because each project can
 # define them, we shouldn't expect to know every possible
 # property-like decorator!
-# TODO(cpopa): just implement descriptors already.
 POSSIBLE_PROPERTIES = {"cached_property", "cachedproperty",
                        "lazyproperty", "lazy_property", "reify",
                        "lazyattribute", "lazy_attribute",
-                       "LazyProperty", "lazy"}
+                       "LazyProperty", "lazy", "cache_readonly"}
 
 
 def _is_property(meth):
@@ -51,11 +59,33 @@ def _is_property(meth):
         return True
     stripped = {name.split(".")[-1] for name in meth.decoratornames()
                 if name is not util.Uninferable}
-    return any(name in stripped for name in POSSIBLE_PROPERTIES)
+    if any(name in stripped for name in POSSIBLE_PROPERTIES):
+        return True
+
+    # Lookup for subclasses of *property*
+    if not meth.decorators:
+        return False
+    for decorator in meth.decorators.nodes or ():
+        inferred = helpers.safe_infer(decorator)
+        if inferred is None or inferred is util.Uninferable:
+            continue
+        if inferred.__class__.__name__ == 'ClassDef':
+            for base_class in inferred.bases:
+                module, _ = base_class.lookup(base_class.name)
+                if module.name == BUILTINS and base_class.name == 'property':
+                    return True
+
+    return False
 
 
-class Proxy(object):
-    """a simple proxy object"""
+class Proxy:
+    """a simple proxy object
+
+    Note:
+
+    Subclasses of this object will need a custom __getattr__
+    if new instance attributes are created. See the Const class
+    """
 
     _proxied = None # proxied object may be set by class or by instance
 
@@ -113,12 +143,15 @@ def _infer_method_result_truth(instance, method_name, context):
     if meth and hasattr(meth, 'infer_call_result'):
         if not meth.callable():
             return util.Uninferable
-        for value in meth.infer_call_result(instance, context=context):
-            if value is util.Uninferable:
-                return value
+        try:
+            for value in meth.infer_call_result(instance, context=context):
+                if value is util.Uninferable:
+                    return value
 
-            inferred = next(value.infer(context=context))
-            return inferred.bool_value()
+                inferred = next(value.infer(context=context))
+                return inferred.bool_value()
+        except exceptions.InferenceError:
+            pass
     return util.Uninferable
 
 
@@ -133,7 +166,7 @@ class BaseInstance(Proxy):
     def getattr(self, name, context=None, lookupclass=True):
         try:
             values = self._proxied.instance_attr(name, context)
-        except exceptions.AttributeInferenceError:
+        except exceptions.AttributeInferenceError as exc:
             if self.special_attributes and name in self.special_attributes:
                 return [self.special_attributes.lookup(name)]
 
@@ -143,9 +176,11 @@ class BaseInstance(Proxy):
                 return self._proxied.getattr(name, context,
                                              class_context=False)
 
-            util.reraise(exceptions.AttributeInferenceError(target=self,
-                                                            attribute=name,
-                                                            context=context))
+            raise exceptions.AttributeInferenceError(
+                target=self,
+                attribute=name,
+                context=context,
+            ) from exc
         # since we've no context information, return matching class members as
         # well
         if lookupclass:
@@ -167,26 +202,25 @@ class BaseInstance(Proxy):
 
             # XXX frame should be self._proxied, or not ?
             get_attr = self.getattr(name, context, lookupclass=False)
-            for stmt in _infer_stmts(self._wrap_attr(get_attr, context),
-                                     context, frame=self):
-                yield stmt
-        except exceptions.AttributeInferenceError:
+            yield from _infer_stmts(self._wrap_attr(get_attr, context), context, frame=self)
+        except exceptions.AttributeInferenceError as error:
             try:
                 # fallback to class.igetattr since it has some logic to handle
                 # descriptors
+                # But only if the _proxied is the Class.
+                if self._proxied.__class__.__name__ != 'ClassDef':
+                    raise exceptions.InferenceError(**vars(error)) from error
                 attrs = self._proxied.igetattr(name, context, class_context=False)
-                for stmt in self._wrap_attr(attrs, context):
-                    yield stmt
+                yield from self._wrap_attr(attrs, context)
             except exceptions.AttributeInferenceError as error:
-                util.reraise(exceptions.InferenceError(**vars(error)))
+                raise exceptions.InferenceError(**vars(error)) from error
 
     def _wrap_attr(self, attrs, context=None):
         """wrap bound methods of attrs in a InstanceMethod proxies"""
         for attr in attrs:
             if isinstance(attr, UnboundMethod):
                 if _is_property(attr):
-                    for inferred in attr.infer_call_result(self, context):
-                        yield inferred
+                    yield from attr.infer_call_result(self, context)
                 else:
                     yield BoundMethod(attr, self)
             elif hasattr(attr, 'name') and attr.name == '<lambda>':
@@ -205,6 +239,7 @@ class BaseInstance(Proxy):
 
     def infer_call_result(self, caller, context=None):
         """infer what a class instance is returning when called"""
+        context = contextmod.bind_context_to_node(context, self)
         inferred = False
         for node in self._proxied.igetattr('__call__', context):
             if node is util.Uninferable or not node.callable():
@@ -271,8 +306,7 @@ class Instance(BaseInstance):
                 return True
         return result
 
-    # TODO(cpopa): this is set in inference.py
-    # The circular dependency hell goes deeper and deeper.
+    # This is set in inference.py.
     def getitem(self, index, context=None):
         pass
 
@@ -289,6 +323,9 @@ class UnboundMethod(Proxy):
                                          self._proxied.name,
                                          frame.qname(), id(self))
 
+    def implicit_parameters(self):
+        return 0
+
     def is_bound(self):
         return False
 
@@ -303,11 +340,25 @@ class UnboundMethod(Proxy):
         return self._proxied.igetattr(name, context)
 
     def infer_call_result(self, caller, context):
+        """
+        The boundnode of the regular context with a function called
+        on ``object.__new__`` will be of type ``object``,
+        which is incorrect for the argument in general.
+        If no context is given the ``object.__new__`` call argument will
+        correctly inferred except when inside a call that requires
+        the additional context (such as a classmethod) of the boundnode
+        to determine which class the method was called from
+        """
+
         # If we're unbound method __new__ of builtin object, the result is an
         # instance of the class given as first argument.
         if (self._proxied.name == '__new__' and
                 self._proxied.parent.frame().qname() == '%s.object' % BUILTINS):
-            infer = caller.args[0].infer() if caller.args else []
+            if caller.args:
+                node_context = context.extra_context.get(caller.args[0])
+                infer = caller.args[0].infer(context=node_context)
+            else:
+                infer = []
             return (Instance(x) if x is not util.Uninferable else x for x in infer)
         return self._proxied.infer_call_result(caller, context)
 
@@ -325,6 +376,9 @@ class BoundMethod(UnboundMethod):
         UnboundMethod.__init__(self, proxy)
         self.bound = bound
 
+    def implicit_parameters(self):
+        return 1
+
     def is_bound(self):
         return True
 
@@ -333,56 +387,52 @@ class BoundMethod(UnboundMethod):
 
         In order for such call to be valid, the metaclass needs to be
         a subtype of ``type``, the name needs to be a string, the bases
-        needs to be a tuple of classes and the attributes a dictionary
-        of strings to values.
+        needs to be a tuple of classes
         """
         from astroid import node_classes
         # Verify the metaclass
         mcs = next(caller.args[0].infer(context=context))
         if mcs.__class__.__name__ != 'ClassDef':
             # Not a valid first argument.
-            return
+            return None
         if not mcs.is_subtype_of("%s.type" % BUILTINS):
             # Not a valid metaclass.
-            return
+            return None
 
         # Verify the name
         name = next(caller.args[1].infer(context=context))
         if name.__class__.__name__ != 'Const':
             # Not a valid name, needs to be a const.
-            return
+            return None
         if not isinstance(name.value, str):
             # Needs to be a string.
-            return
+            return None
 
         # Verify the bases
         bases = next(caller.args[2].infer(context=context))
         if bases.__class__.__name__ != 'Tuple':
             # Needs to be a tuple.
-            return
+            return None
         inferred_bases = [next(elt.infer(context=context))
                           for elt in bases.elts]
         if any(base.__class__.__name__ != 'ClassDef'
                for base in inferred_bases):
             # All the bases needs to be Classes
-            return
+            return None
 
         # Verify the attributes.
         attrs = next(caller.args[3].infer(context=context))
         if attrs.__class__.__name__ != 'Dict':
             # Needs to be a dictionary.
-            return
+            return None
         cls_locals = collections.defaultdict(list)
         for key, value in attrs.items:
             key = next(key.infer(context=context))
             value = next(value.infer(context=context))
-            if key.__class__.__name__ != 'Const':
-                # Something invalid as an attribute.
-                return
-            if not isinstance(key.value, str):
-                # Not a proper attribute.
-                return
-            cls_locals[key.value].append(value)
+            # Ignore non string keys
+            if (key.__class__.__name__ == 'Const' and
+                    isinstance(key.value, str)):
+                cls_locals[key.value].append(value)
 
         # Build the class from now.
         cls = mcs.__class__(name=name.value, lineno=caller.lineno,
@@ -395,19 +445,12 @@ class BoundMethod(UnboundMethod):
         return cls
 
     def infer_call_result(self, caller, context=None):
-        if context is None:
-            context = contextmod.InferenceContext()
-        context = context.clone()
-        context.boundnode = self.bound
-
+        context = contextmod.bind_context_to_node(context, self.bound)
         if (self.bound.__class__.__name__ == 'ClassDef'
                 and self.bound.name == 'type'
                 and self.name == '__new__'
-                and len(caller.args) == 4
-                # TODO(cpopa): this check shouldn't be needed.
-                and self._proxied.parent.frame().qname() == '%s.object' % BUILTINS):
-
-            # Check if we have an ``type.__new__(mcs, name, bases, attrs)`` call.
+                and len(caller.args) == 4):
+            # Check if we have a ``type.__new__(mcs, name, bases, attrs)`` call.
             new_cls = self._infer_type_new_call(caller, context)
             if new_cls:
                 return iter((new_cls, ))
