@@ -1,7 +1,14 @@
 # Copyright (c) 2006-2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
-# Copyright (c) 2014-2016 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2012 FELD Boris <lothiraldan@gmail.com>
+# Copyright (c) 2014-2018 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2014 Google, Inc.
-# Copyright (c) 2015-2016 Cara Vinson <ceridwenv@gmail.com>
+# Copyright (c) 2015-2016 Ceridwen <ceridwenv@gmail.com>
+# Copyright (c) 2015 Florian Bruhin <me@the-compiler.org>
+# Copyright (c) 2015 Ovidiu Sabou <ovidiu@sabou.org>
+# Copyright (c) 2016 Derek Gustafson <degustaf@gmail.com>
+# Copyright (c) 2016 Jakub Wilk <jwilk@jwilk.net>
+# Copyright (c) 2018 Nick Drozd <nicholasdrozd@gmail.com>
+# Copyright (c) 2018 Bryce Guinta <bryce.paul.guinta@gmail.com>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -10,14 +17,13 @@
 (build_* functions) or from living object (object_build_* functions)
 """
 
+import builtins
 import inspect
 import logging
 import os
 import sys
 import types
 
-import six
-
 from astroid import bases
 from astroid import manager
 from astroid import node_classes
@@ -26,9 +32,9 @@ from astroid import nodes
 
 MANAGER = manager.AstroidManager()
 # the keys of CONST_CLS eg python builtin types
+
 _CONSTANTS = tuple(node_classes.CONST_CLS)
-_JYTHON = os.name == 'java'
-_BUILTINS = vars(six.moves.builtins)
+_BUILTINS = vars(builtins)
 _LOG = logging.getLogger(__name__)
 
 
@@ -88,7 +94,7 @@ def attach_import_node(node, modname, membername):
 
 
 def build_module(name, doc=None):
-    """create and initialize a astroid Module node"""
+    """create and initialize an astroid Module node"""
     node = nodes.Module(name, doc, pure_python=False)
     node.package = False
     node.parent = None
@@ -96,7 +102,7 @@ def build_module(name, doc=None):
 
 
 def build_class(name, basenames=(), doc=None):
-    """create and initialize a astroid ClassDef node"""
+    """create and initialize an astroid ClassDef node"""
     node = nodes.ClassDef(name, doc)
     for base in basenames:
         basenode = nodes.Name()
@@ -107,7 +113,7 @@ def build_class(name, basenames=(), doc=None):
 
 
 def build_function(name, args=None, defaults=None, doc=None):
-    """create and initialize a astroid FunctionDef node"""
+    """create and initialize an astroid FunctionDef node"""
     args, defaults = args or [], defaults or []
     # first argument is now a list of decorators
     func = nodes.FunctionDef(name, doc)
@@ -221,7 +227,7 @@ def _base_class_object_build(node, member, basenames, name=None, localname=None)
 def _build_from_function(node, name, member, module):
     # verify this is not an imported function
     try:
-        code = six.get_function_code(member)
+        code = member.__code__
     except AttributeError:
         # Some implementations don't provide the code object,
         # such as Jython.
@@ -236,7 +242,7 @@ def _build_from_function(node, name, member, module):
         object_build_function(node, member, name)
 
 
-class InspectBuilder(object):
+class InspectBuilder:
     """class for building nodes from living object
 
     this is actually a really minimal representation, including only Module,
@@ -285,7 +291,7 @@ class InspectBuilder(object):
                 attach_dummy_node(node, name)
                 continue
             if inspect.ismethod(member):
-                member = six.get_method_function(member)
+                member = member.__func__
             if inspect.isfunction(member):
                 _build_from_function(node, name, member, self._module)
             elif inspect.isbuiltin(member):
@@ -321,6 +327,7 @@ class InspectBuilder(object):
             else:
                 # create an empty node so that the name is actually defined
                 attach_dummy_node(node, name, member)
+        return None
 
     def imported_member(self, node, member, name):
         """verify this is not an imported class or handle it"""
@@ -334,12 +341,11 @@ class InspectBuilder(object):
                            'astroid from living object')
             modname = None
         if modname is None:
-            if (name in ('__new__', '__subclasshook__')
-                    or (name in _BUILTINS and _JYTHON)):
+            if name in ('__new__', '__subclasshook__'):
                 # Python 2.5.1 (r251:54863, Sep  1 2010, 22:03:14)
                 # >>> print object.__new__.__module__
                 # None
-                modname = six.moves.builtins.__name__
+                modname = builtins.__name__
             else:
                 attach_dummy_node(node, name, member)
                 return True
@@ -371,7 +377,6 @@ def _astroid_bootstrapping(astroid_builtin=None):
     # this boot strapping is necessary since we need the Const nodes to
     # inspect_build builtins, and then we can proxy Const
     if astroid_builtin is None:
-        from six.moves import builtins
         astroid_builtin = Astroid_BUILDER.inspect_build(builtins)
 
     # pylint: disable=redefined-outer-name
@@ -392,18 +397,16 @@ def _astroid_bootstrapping(astroid_builtin=None):
 _astroid_bootstrapping()
 
 # TODO : find a nicer way to handle this situation;
-# However __proxied introduced an
-# infinite recursion (see https://bugs.launchpad.net/pylint/+bug/456870)
 def _set_proxied(const):
     return _CONST_PROXY[const.value.__class__]
 nodes.Const._proxied = property(_set_proxied)
 
 _GeneratorType = nodes.ClassDef(types.GeneratorType.__name__, types.GeneratorType.__doc__)
-_GeneratorType.parent = MANAGER.astroid_cache[six.moves.builtins.__name__]
+_GeneratorType.parent = MANAGER.astroid_cache[builtins.__name__]
 bases.Generator._proxied = _GeneratorType
 Astroid_BUILDER.object_build(bases.Generator._proxied, types.GeneratorType)
 
-_builtins = MANAGER.astroid_cache[six.moves.builtins.__name__]
+_builtins = MANAGER.astroid_cache[builtins.__name__]
 BUILTIN_TYPES = (types.GetSetDescriptorType, types.GeneratorType,
                  types.MemberDescriptorType, type(None), type(NotImplemented),
                  types.FunctionType, types.MethodType,
@@ -411,6 +414,6 @@ BUILTIN_TYPES = (types.GetSetDescriptorType, types.GeneratorType,
 for _type in BUILTIN_TYPES:
     if _type.__name__ not in _builtins:
         cls = nodes.ClassDef(_type.__name__, _type.__doc__)
-        cls.parent = MANAGER.astroid_cache[six.moves.builtins.__name__]
+        cls.parent = MANAGER.astroid_cache[builtins.__name__]
         Astroid_BUILDER.object_build(cls, _type)
         _builtins[_type.__name__] = cls
